using System;
using System.Collections.Generic;
using System.Threading;
using NDesk.DBus;

using Nemo;

[Interface("org.freedesktop.Tracker.Metadata")]
public interface Metadata
{
    string[] GetRegisteredTypes(string metadata_class);
	void RegisterType(string name, string datatype);
    string[] Get(string service, string id, string[] keys);
//    void Set(string service, string id, string[] keys, string[] values);
}

[Interface("org.freedesktop.Tracker.Search")]
public interface Search
{
//    string[] Text(int live_query_id, string service, string search_text, int offset, int max_hits);
    string GetSnippet(string service, string id, string search_text);
    string[][] Query(int live_query_id, string service, string[] fields, string search_text, string keywords,
                     string query_condition, bool sort_by_service, int offset, int max_hits);
    string[][] TextDetailed(int live_query_id, string service, string search_text, int offset, int max_hits);
}

[Interface("org.freedesktop.Tracker")]
public interface ITracker
{
    int GetVersion();
}

public sealed class Broker
{
    private ITracker bus;
    private Search bus_search;
    private Metadata bus_metadata;
    private Connection con;

	MetadataStore meta;	
	
    private bool iterating;

    private string[] registered_fields;

    private Broker()
    {
		// tracker specific
		
		finished = false;
		
		cur_search_id = 0;

		callbacks = new List<VoidFunction>();

        con = Bus.Session;

		//string addr = "unix:path=/tmp/dbus-ABCDEFGHIJ";
        //con = new Connection (addr);

        ObjectPath opath = new ObjectPath("/org/freedesktop/tracker");

        bus = con.GetObject<ITracker>("org.freedesktop.Tracker", opath);
        bus_search = con.GetObject<Search>("org.freedesktop.Tracker", opath);
        bus_metadata = con.GetObject<Metadata>("org.freedesktop.Tracker", opath);

        registered_fields = new string[] { "File:Path", "File:Modified", "File:Accessed", "File:Size", "File:Name", "File:Mime" }; // "File:Rank", 

        start_iterating();

		// metadata store
		if (meta == null) {
			meta = new MetadataStore(Singleton<Configuration>.Instance.data.watch_dir);
        	meta.start(System.IO.Path.Combine(Environment.GetFolderPath(System.Environment.SpecialFolder.Personal), ".nemo/"));
        }
	}

	bool finished;

    public void start_iterating() 
    {
        if (iterating) // re-entrance protection
            return;

		ThreadPool.QueueUserWorkItem(delegate(Object stateInfo) {
            while (!finished)
                con.Iterate();
		});

		own_thread_pool();

        iterating = true;
    }

    public void stop() 
    {
        meta.stop();
        finished = true;
        bus.GetVersion(); // final iteration
        iterating = false;
    }

	public void reindex_metadata_store()
	{
	    System.Console.WriteLine("Reindexing filesystem!");

		if (meta == null) {
			meta = new MetadataStore(Singleton<Configuration>.Instance.data.watch_dir);
        	meta.start(System.IO.Path.Combine(Environment.GetFolderPath(System.Environment.SpecialFolder.Personal), ".nemo/"));
        }

		System.Console.WriteLine(Singleton<Configuration>.Instance.data.watch_dir);

        meta.sync_with_filesystem();
	}

 	public int RegisterItemType(string[] fields) 
	{
		string[] new_registered_fields = new string[fields.Length + registered_fields.Length]; 
		
		int i = 0;
		
		foreach (string field in registered_fields)
		{
			new_registered_fields[i++] = field;
		}
		
		foreach (string field in fields)
		{
			new_registered_fields[i++] = field;
		}
		
		int registered_index = registered_fields.Length;  
		
		registered_fields = new_registered_fields;
		
		return registered_index;
	}
 
 	public void get_metadata_from_tracker(string filename, VoidFunction<string[]> callback)
 	{
	 	ThreadPool.QueueUserWorkItem(delegate {
 			callback(bus_metadata.Get("Files", filename, registered_fields));
 		});
 	}
 
 	public void get_file_from_metadata_store(string filename, VoidFunction<File> callback)
 	{
	 	ThreadPool.QueueUserWorkItem(delegate {
 			meta.get_file_from_path(filename, callback);
 		});
 	}
 
 	private int cur_search_id;

	// result returned as array of [type, uri]
	public void SearchForText(string search_string, bool fast, VoidFunction<List<Tuple<string[], File>>> callback)
	{
		ThreadPool.QueueUserWorkItem(delegate {
            ++cur_search_id;

            int search_id = cur_search_id;

            if (!fast)
                Thread.Sleep(1000);

            if (cur_search_id == search_id)
            {
				// the interface for searching using tracker is borken so we need to get 500 to make sure we get some we like...
				string[][] result = bus_search.TextDetailed(-1, "Files", search_string, 0, 500);

				if (cur_search_id == search_id) {

					List<string> filenames = new List<string>();

					foreach (string[] r in result) {

						System.Console.WriteLine("{0} - {1}", r[0], r[1]);
			
						if (r[1] == "Folders") // sigh, skip folders
							continue;

						filenames.Add(r[0]);
					}
					
					System.Console.WriteLine("search metadata store for {0} files", filenames.Count);
					
					meta.get_files_from_paths(filenames, Helpers.RunInMainThread<List<File>>(delegate (List<File> files) {

						System.Console.WriteLine("got {0} files from metadata store", files.Count);

						List<Tuple<string[], File>> real_results = new List<Tuple<string[], File>>();

						foreach (File file in files)
							real_results.Add(new Tuple<string[], File>(bus_metadata.Get("Files", file.path, registered_fields), file));

						callback(real_results);
					}));
				}
            }
        });
	}
	
    public void Search(List<MetaLabel> keywords, List<long> types, DateTime date_start, DateTime date_end, bool starred, 
    				   VoidFunction<List<File>> callback)
    {
		Query q = new Query();
		q.labels = keywords;
		q.types = types;
		q.date_start = date_start;
		q.date_end = date_end;
		q.starred = starred;

#if false
		meta.set_query_result_callback(delegate (List<File> files) { 
			meta.set_query_result_callback(null); 
			callback(files);
		});

		meta.set_query_result_callback(callback);
		
		meta.set_query(q);
#endif		

        ++cur_search_id;

        int search_id = cur_search_id;

		meta.search(q, delegate(List<File> files) { 
			if (cur_search_id == search_id)
				callback(files);
		});

//        meta.trigger_query_result_changed();
    }

    public void get_all_labels(VoidFunction<List<MetaLabel>> callback)
    {
		meta.set_labels_changed_callback(callback);
		meta.trigger_labels_changed();
    }

    public void get_type_labels(VoidFunction<List<Tuple<FileTypeCategory, int>>> callback)
    {
		meta.set_type_labels_changed_callback(callback);
		meta.trigger_type_labels_changed();
    }

	private void own_thread_pool()
	{
		ThreadPool.QueueUserWorkItem(delegate(Object stateInfo) {
			while (true) {
				System.Threading.Monitor.Enter(callback_lock);
				System.Threading.Monitor.Wait(callback_lock);
				if (callbacks.Count > 0) {
					callbacks[0]();
					callbacks.RemoveAt(0);
				}
				System.Threading.Monitor.Exit(callback_lock);
			}
		});
	}

	// own threadpool implementation because the one in mono is broken
    private static object callback_lock = new Object();
	private List<VoidFunction> callbacks;

    public void add_label(long parent_id, string label)
    {
        MetaLabel new_label = new MetaLabel();
        new_label.label = label;
        new_label.parent_db_id = parent_id;

		meta.add_label(new_label, delegate {});
    }	
	
	public void remove_label(MetaLabel label)
    {
		meta.remove_label(label, delegate {});
    }
    
	public void rename_label(MetaLabel label, string new_name)
	{
		meta.rename_label(label, new_name);
	}
	
    public void add_label_to_file(File file, MetaLabel label)
	{
        meta.add_label_to_file(file, label, delegate {});		
    }

	public void remove_label_from_file(File file, MetaLabel label)
	{
        meta.remove_label_from_file(file, label, delegate {});
    }
		
	public void set_starred(File file, bool starred)
	{
	    meta.set_starred(file, starred, delegate {});	
    }	
	
    public void search_snippet(string uri, string search_string, VoidFunction<string> callback)
    {
		ThreadPool.QueueUserWorkItem(delegate {
			string result = "";
			if (search_string.Length != 0)
				result = bus_search.GetSnippet("Files", uri, search_string);
			callback(result);
		});
    }

    public void get_text(string uri, VoidFunction<string> callback)
    {
    	// FIXME: remember thread pool
//    	System.Console.WriteLine("TRACKER IS FUBARED");
		//string result = bus_files.GetTextContents(uri, 0, 100);
		callback(""); // result
    }
} 
